#include "catch.hpp"

#include "small1.hpp"

#include <string>
#include <sstream>
#include <iostream>


TEST_CASE("parse_matrix", "[small1]") {

    std::stringstream str;
    str.exceptions(std::ios_base::badbit);

    GIVEN("Easy to parse rectangular matrix") {
        std::string s = R"f(1 2 3 4
5 6 7 8
9 10 11 12
13 14 15 16
)f";

        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are equal and correct") {
                REQUIRE(temp.first == 4);
                REQUIRE(temp.second == 4);
            }
        }
    }

    GIVEN("Singular column matrix") {
        std::string s = "1\n2\n3\n4\n5\n6\n7\n";

        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are correct") {
                REQUIRE(temp.first == 7);
                REQUIRE(temp.second == 1);
            }
        }
    }

    GIVEN("Singular row matrix") {
        std::string s = "1 2 3 4 5 6 7\n";

        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are correct") {
                REQUIRE(temp.first == 1);
                REQUIRE(temp.second == 7);
            }
        }
    }


    GIVEN("Wide matrix") {
        std::string s = R"f(1 2 3 4 5 6 7 8
1 2 3 4 5 6 7 8
1 2 3 4 5 6 7 8
)f";

        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are correct") {
                REQUIRE(temp.first == 3);
                REQUIRE(temp.second == 8);
            }
        }
    }

    GIVEN("Empty string (no matrix)") {
        std::string s = "";
        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are equal to zero") {
                REQUIRE(temp.first == 0);
                REQUIRE(temp.second == 0);
            }
        }
    }

    GIVEN("String with single number") {
        std::string s = "11\n";
        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are equal to one") {
                REQUIRE(temp.first == 1);
                REQUIRE(temp.second == 1);
            }
        }
    }

    GIVEN("Tall matrix") {
        std::string s = R"f(1 2 3
4 5 6
7 8 9
10 11 12
13 14 15
)f";

        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are correct") {
                REQUIRE(temp.first == 5);
                REQUIRE(temp.second == 3);
            }
        }
    }

    GIVEN("Matrix with weird whitespace") {
        std::string s = R"f(1 2		3 4
5		6		7 8
9	10 12 13
13 14		15 33
16 17	22	11  
)f";
        str.str(s);

        WHEN("matrix is parsed") {
            auto temp = parse_matrix(str);
            THEN("dimensions are correct") {
                REQUIRE(temp.first == 5);
                REQUIRE(temp.second == 4);
            }
        }
    }

    GIVEN("Matrix with rows that have different width") {
        std::string s = R"f(1 2 3 4
5 6 7
)f";
        str.str(s);

        WHEN("matrix is parsed") {
            THEN("exception is thrown (inconsistent row sizes)") {
                REQUIRE_THROWS_AS(auto temp = parse_matrix(str);, std::invalid_argument);
            }
        }
    }

    GIVEN("Matrix with middle row different") {
        std::string s = R"f(1 2 3 4
4 56 7 8
5 6 7
77 889 97 10
8 9 10 11
)f";
        str.str(s);

        WHEN("matrix is parsed") {
            THEN("exception is thrown (inconsistent row sizes)") {
                REQUIRE_THROWS_AS(auto temp = parse_matrix(str);, std::invalid_argument);
            }
        }
    }
}

TEST_CASE("print_table", "[small1]") {
    std::stringstream str;

    SECTION("input 1") {
        std::vector<std::string> vec = {"abc", "def", "ghi", "jkl"};
        print_table(str, vec);
        std::string exp = 
R"f(-----------
| abc| def|
| ghi| jkl|
-----------
)f";
        REQUIRE(str.str() == exp);
    }
    
    SECTION("input 2") {
        std::vector<std::string> vec = {"123", "456", "789", "012"};
        print_table(str, vec);
        std::string exp = 
R"f(-----------
| 123| 456|
| 789| 012|
-----------
)f";
        REQUIRE(str.str() == exp);
    }
    
    SECTION("input 3") {
        std::vector<std::string> vec = {"abcd", "def", "ghi", "jkl"};
        print_table(str, vec);
        std::string exp = 
R"f(-------------
| abcd|  def|
|  ghi|  jkl|
-------------
)f";
        REQUIRE(str.str() == exp);
    }
    
    SECTION("input 4") {
        std::vector<std::string> vec = {"abcd", "def", "ghi"};
        print_table(str, vec);
        std::string exp = 
R"f(-------------
| abcd|  def|
|  ghi|     |
-------------
)f";
        REQUIRE(str.str() == exp);
    }
    
    SECTION("input 5") {
        std::vector<std::string> vec = {
            "James", "David", "Christopher", "George", "Ronald", "John", "Richard",
            "Daniel", "Kenneth", "Anthony", "Robert", "Charles", "Paul", "Steven",
            "Kevin", "Michael", "Joseph", "Mark", "Edward", "Jason", "William", "Thomas",
            "Donald", "Brian"
        };
        print_table(str, vec);
        std::string exp = 
R"f(---------------------------
|       James|       David|
| Christopher|      George|
|      Ronald|        John|
|     Richard|      Daniel|
|     Kenneth|     Anthony|
|      Robert|     Charles|
|        Paul|      Steven|
|       Kevin|     Michael|
|      Joseph|        Mark|
|      Edward|       Jason|
|     William|      Thomas|
|      Donald|       Brian|
---------------------------
)f";
        REQUIRE(str.str() == exp);
    }
}

TEST_CASE("Line validation", "[small1]") {

    GIVEN("properly tagged equal numbers") {
        CHECK(validate_line("true 2 2 2"));
    }

    GIVEN("improperly tagged equal string") {
        CHECK(!validate_line("false 2 2 2"));
    }

    GIVEN("properly tagged non-equal string") {
        CHECK(validate_line("false 2 3 3"));
    }

    GIVEN("improperly tagged non-equal string") {
        CHECK(!validate_line("true 2 3 3"));
    }

    GIVEN("complex properly tagged equal string") {
        CHECK(validate_line("true 0111 73 0x49"));
    }

    GIVEN("complex improperly tagged equal string") {
        CHECK(!validate_line("false 0111 73 0x49"));
    }

    GIVEN("complex properly tagged non-equal string") {
        CHECK(validate_line("false 0131 74 0x4A"));
    }

    GIVEN("complex improperly tagged non-equal string") {
        CHECK(!validate_line("false 07756 4078 0xFEE"));
    }

    GIVEN("string not conforming to format -- missing values") {
        CHECK(!validate_line("false 112"));
    }

    GIVEN("string not conforming to format -- bad number format") {
        CHECK(!validate_line("false 88 88 0x58"));
    }

    GIVEN("string not conforming to format -- not numbers") {
        CHECK(!validate_line("false 07 7 G"));
    }
}

TEST_CASE("Number comparison", "[small1]") {
    std::stringstream sstream;
    GIVEN("Empty stream") {
        CHECK(max_number(sstream) == "");
    }
    GIVEN("Single number in stream") {
        sstream << "123";
        CHECK(max_number(sstream) == "123");
    }
    GIVEN("Same length numbers in stream") {
        sstream << "123,345,677,231";
        CHECK(max_number(sstream) == "677");
    }

    GIVEN("Different length numbers in stream") {
        sstream << "1,22,333,444,4444";
        CHECK(max_number(sstream) == "4444");
    }

    GIVEN("Same length looong numbers") {
        sstream << "12345678909876543211234567890,"
                << "12345678909876543212234567890";
        CHECK(max_number(sstream) == "12345678909876543212234567890");
    }

    GIVEN("Different length loooong numbers") {
        sstream << "123456789098765432123456789,"
                << "12345678909876543212345678,"
                << "1234567890987654321234567,"
                << "1234567890987654321234567890";
        CHECK(max_number(sstream) == "1234567890987654321234567890");
    }
}
